package com.divillysausages.sprint.anim 
{
	import com.divillysausages.ds.asset.Asset;
	import com.divillysausages.ds.collections.objectpool.IPoolObj;
	import com.divillysausages.ds.DS;
	import flash.display.Bitmap;
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.display.MovieClip;
	import flash.geom.Point;
	
	/**
	 * An animation that we're going to play
	 * @author Damian Connolly
	 */
	public class Anim implements IPoolObj
	{
		
		/*************************************************************************************************************/
		
		/**
		 * The default anim for any movieclip (mostly from 1 - totalFrames)
		 */
		public static const DEFAULT:String = "default";
		
		/*************************************************************************************************************/
		
		/**
		 * The previous anim in this linked list
		 */
		public var prev:Anim = null;
		
		/**
		 * The next anim in this linked list
		 */
		public var next:Anim = null;
		
		/**
		 * Is this an idle animation? Idle animations are looped animations that play when you're not playing
		 * any other animation
		 */
		public var isIdleAnim:Boolean = false;
		
		/*************************************************************************************************************/
		
		private var m_calledLoadGraphics:Boolean		= false;// if we need to load some graphics, are they good?
		private var m_isCurrent:Boolean					= false;// is this anim the current anim (i.e. trying to play, though we might be waiting on graphics)
		private var m_isInited:Boolean					= false;// are we inited yet?
		private var m_isPaused:Boolean					= false;// is the anim paused?
		private var m_isFinished:Boolean				= false;// is the anim finished?
		private var m_target:DisplayObject 				= null; // the target for this animation
		private var m_params:AnimParams					= null; // the anim params for this anim
		private var m_dObjAddTo:DisplayObjectContainer	= null;	// the container we should add ourselves to (used when we call, but we don't have our graphics yet)
		private var m_dObjAddToPos:Point				= null; // the position to add ourselves at
		private var m_dObjAdd:DisplayObject				= null; // the display object to add to us (used when we call, but we don't have our graphics yet)
		private var m_dObjAddInstance:String			= null; // the instance name (if any) of the clip where we want to add the display object
		private var m_dObjAddInstancePos:Point			= null; // the position to add the instance at
		private var m_currTime:Number				 	= 0.0;	// the current time
		private var m_tickTime:Number					= 0.0;	// how much time we have for a tick
		private var m_frames:Vector.<AnimFrameInfo>		= null; // any frames that we have for this animation
		private var m_currFrame:int						= 0;	// our current frame
		private var m_prevFrame:int						= 0;	// our previous frame
		private var m_prevFrameTick:int					= 0;	// our previous frame (in the last tick - frames can last more than one tick so this is used for onUpdate + onFrame)
		
		/*************************************************************************************************************/
		
		/**
		 * The ID for the anim
		 */
		public function get id():String { return ( this.m_params != null ) ? this.m_params.id : null; }
		
		/**
		 * The target for this animation. Either a MovieClip or a Bitmap, or null, if our graphics haven't loaded yet
		 */
		public function get target():DisplayObject { return this.m_target; }
		
		/**
		 * The animation parameters for this anim
		 */
		public function get params():AnimParams { return this.m_params; }
		public function set params( a:AnimParams ):void
		{
			// set our anim params
			this.m_params = a;
			
			// if we have a swf and instance, set them to loading
			if ( this.m_params.target == null && this.m_params.targetClass == null && !this.m_calledLoadGraphics )
			{
				var asset:Asset = DS.assetManager.getAsset( this.m_params.swf );
				if ( asset != null )
					this._onSWFLoad( asset );
				else
					this._onSWFFail();
				this.m_calledLoadGraphics = true;
			}
		}
		
		/**
		 * The current frame of our animation
		 */
		public function get currFrame():int { return this.m_currFrame; }
		public function set currFrame( i:int ):void
		{
			// make sure it's within bounds
			if ( this.m_params != null )
				i = ( i < this.m_params.start ) ? this.m_params.start : ( i > this.m_params.end ) ? this.m_params.end : i;
				
			// set our current frame
			this.m_currFrame = i;
			
			// set our previous frame values
			this.m_prevFrame		= this.m_currFrame - 1;
			this.m_prevFrameTick	= this.m_currFrame - 1;
			
			// reset our times
			this.m_currTime = 0.0;
		}
		
		/**
		 * The parent of our target, if we have one
		 */
		public function get parent():DisplayObjectContainer { return ( this.m_target != null ) ? this.m_target.parent : null; }
		
		/**
		 * Is this the current animation. isCurrent is not exactly the same as isPlaying, as it's possible that you're
		 * waiting for graphics to load or that the anim is paused
		 */
		public function get isCurrent():Boolean { return this.m_isCurrent; }
		
		/**
		 * Has the animation been inited yet?
		 */
		public function get isInited():Boolean { return this.m_isInited; }
		
		/**
		 * Is the animation currently paused?
		 */
		public function get isPaused():Boolean { return this.m_isPaused; }
		
		/**
		 * Is the animation currently playing? (graphics are loaded and we're animating)
		 */
		public function get isPlaying():Boolean { return ( this.m_isCurrent && !this.m_isPaused ); }
		
		/**
		 * Is the animation finished?
		 */
		public function get isFinished():Boolean { return this.m_isFinished; }
		
		/*************************************************************************************************************/
		
		/**
		 * The AnimFrameInfo objects for this anim
		 */
		internal function set frames( v:Vector.<AnimFrameInfo> ):void
		{
			this.m_frames = v;
		}
		
		/*************************************************************************************************************/
		
		/**
		 * Inits the anim
		 */
		public function init( params:* = null ):void 
		{
			// we're now the current anim. Note that the current anim doesn't necessarily
			// mean that you're playing, as you might be waiting on graphics
			this.m_isCurrent = true;
			
			// make sure our params are set
			if ( this.m_params == null )
			{
				DS.error( this, "Can't init the anim, as we don't have a params object" );
				this.pause();
				return;
			}
			
			// make sure we either have a target or a class
			if ( this.m_params.target == null && this.m_params.targetClass == null )
			{
				// pause us while are graphics get loaded
				if ( !this.m_calledLoadGraphics )
				{
					var asset:Asset = DS.assetManager.getAsset( this.m_params.swf );
					if ( asset != null )
						this._onSWFLoad( asset );
					else
						this._onSWFFail();
					this.m_calledLoadGraphics = true;
				}
				this.pause();
				return;
			}
			
			// store our target
			if ( this.m_params.target != null )
				this.m_target = this.m_params.target;
			else
			{
				this.m_target = new this.m_params.targetClass;
				
				// make sure it's a movieclip or a bitmap
				if ( !( this.m_target is MovieClip ) && !( this.m_target is Bitmap ) )
				{
					DS.error( this, "The target class (" + this.m_params.targetClass + ") isn't a MovieClip or a Bitmap - won't animate" );
					this.m_target	= null;	// clear the target and pause
					this.pause();
					return;
				}
			}
			
			// if we're a bitmap, make sure we have some bmds to show
			if ( this.m_target is Bitmap && this.m_params.bmds == null )
			{
				DS.error( this, "A Bitmap animation requires a Vector of BitmapData to play" );
				this.m_target = null; // clear the target and pause
				this.pause();
				return;
			}
			
			// set our position from the params (or set it)
			var isSetX:Boolean = !this._isNaN( this.m_params.x );
			var isSetY:Boolean = !this._isNaN( this.m_params.y );
			if ( isSetX || isSetY || this.m_params.bmdOffsets != null )
				this._setPos( true );
			else
			{
				// we need to set the params to not NaN for this to work
				if ( !isSetX ) this.m_params.x = 0.0;
				if ( !isSetY ) this.m_params.y = 0.0;
				this._setPos( false );
			}
				
			// set our scale from the params
			this.m_target.scaleX = ( !this._isNaN( this.m_params.scaleX ) ) ? this.m_params.scaleX : this.m_target.scaleX;
			this.m_target.scaleY = ( !this._isNaN( this.m_params.scaleY ) ) ? this.m_params.scaleY : this.m_target.scaleY;
			
			// if we have a parent in our params (we should), add us
			if ( this.m_params.parent != null )
			{
				if( this.m_params.childIndex == -1 || this.m_params.childIndex > this.m_params.parent.numChildren )
					this.m_params.parent.addChild( this.m_target ); // just do a normal add child
				else
					this.m_params.parent.addChildAt( this.m_target, this.m_params.childIndex ); // add it at the right index
			}
				
			// set our inital tick time
			if ( this.m_params.tps > 0 )
				this.m_tickTime = int( 10000 / this.m_params.tps ) * 0.0001; // rounds it to 4 decimal places
				
			// set our loops to infinite if we're an idle anim
			if ( this.isIdleAnim )
				this.m_params.loops = -1;
				
			// if we don't have a start and end, get them
			if ( this.m_target is MovieClip )
			{
				// stop it from playing
				var mc:MovieClip = this.m_target as MovieClip;
				mc.gotoAndStop( 0 );
				
				// if we have an animFrameInfo, set the start and end of the params
				if ( this.m_params.animFrameInfo != null )
				{
					// if it's a string, get our an frame obj
					var name:String = null;
					if ( this.m_params.animFrameInfo is String )
					{
						name						= this.m_params.animFrameInfo;
						this.m_params.animFrameInfo = this.getFrameInfo( this.m_params.animFrameInfo );
					}
						
					// if we have one, set the start and end
					if ( this.m_params.animFrameInfo != null )
					{
						this.m_params.start = this.m_params.animFrameInfo.start;
						this.m_params.end	= this.m_params.animFrameInfo.end;
					}
					else
						DS.warn( this, "Couldn't find frame info with name '" + name + "', so couldn't set the start or end frames" );
				}
				
				// if we need to get our start and end frames, do it
				if ( this.m_params.start == 0 && this.m_params.end == 0 )
				{
					this.m_params.start = mc.currentFrame;
					this.m_params.end	= mc.totalFrames;
				}
				
				// if we have a loopTo as a string, find the AnimFrameInfo and set the loop
				if ( this.m_params.loopTo != null && this.m_params.loopTo is String )
				{
					var info:AnimFrameInfo = this.getFrameInfo( this.m_params.loopTo );
					if ( info != null )
					{
						// get the loop to frame from our AnimFrameInfo and check to make sure it's good
						this.m_params.loopTo = info.start;
						if ( this.m_params.loopTo < this.m_params.start || this.m_params.loopTo > this.m_params.end )
						{
							DS.warn( this, "The loop to that we have (" + this.m_params.loopTo + ") is outside the bounds of our start (" + this.m_params.start + ") and end (" + this.m_params.end + "), so we can't use it" );
							this.m_params.loopTo = null;
						}
					}
					else
						DS.warn( this, "Couldn't find frame info with name '" + this.m_params.loopTo + "' so we can't do a custom loop" );
				}
			}
			else
			{
				// it's a bitmap, so set the first bitmap data
				( this.m_target as Bitmap ).bitmapData = this.m_params.bmds[0];
				
				// if we need to get our start and end frames, do it
				if ( this.m_params.start == 0 && this.m_params.end == 0 )
				{
					this.m_params.start = 0;
					this.m_params.end	= this.m_params.bmds.length - 1;
				}
			}
			
			// set the current frame
			this.currFrame = this.m_params.start;
			
			// if we're paused, resume() us. we can be paused if we're waiting for our graphics to load
			if ( this.m_isPaused )
				this.resume();
			
			// we're inited
			this.m_isInited = true;
			
			// if we have to be added to something, do it here
			var x:Number = NaN;
			var y:Number = NaN;
			if ( this.m_dObjAddTo != null )
			{
				// position us if needed
				if ( this.m_dObjAddToPos != null )
				{
					x = this.m_dObjAddToPos.x;
					y = this.m_dObjAddToPos.y;
				}
				this.addToParent( this.m_dObjAddTo, x, y );
			}
			
			// if we have something to add, do it here
			x = NaN;
			y = NaN;
			if ( this.m_dObjAdd != null )
			{
				// position them if needed
				if ( this.m_dObjAddInstancePos != null )
				{
					x = this.m_dObjAddInstancePos.x;
					y = this.m_dObjAddInstancePos.y;
				}
				this.addChild( this.m_dObjAdd, this.m_dObjAddInstance, x, y );
			}
			
			// if we have an onInit function, call it
			if ( this.m_params.onInit != null )
			{
				// concat up our args, and call our function (figure out if we're passing the anim or not)
				var numArgs:int	= ( this.m_params.onInitParams != null ) ? this.m_params.onInitParams.length : 0;
				var args:Array 	= ( this.m_params.onInit.length > numArgs ) ? [this] : null;
				if ( this.m_params.onInitParams != null )
					args = ( args == null ) ? this.m_params.onInitParams : args.concat( this.m_params.onInitParams );
				this.m_params.onInit.apply( null, args );
			}
		}
		
		/**
		 * Destroys the anim object and clears it for garbage collection
		 */
		public function destroy():void 
		{
			this.reset();
		}
		
		/**
		 * Soft resets the anim object, but allowing the anim to be reused
		 */
		public function softReset():void
		{				
			// if we have a target, remove it from it's parent (and a parent is specified in the params)
			if ( this.m_target != null && this.m_target.parent != null && this.m_params.parent != null )
				this.m_target.parent.removeChild( this.m_target );
				
			// reset any bools
			this.m_isCurrent	= false;
			this.m_isInited 	= false;
			this.m_isPaused		= false;
			this.m_isFinished	= false;
			
			// reset any times/frames etc
			this.m_currFrame 		= 0;
			this.m_prevFrame		= 0;
			this.m_prevFrameTick	= 0;
			this.m_currTime			= 0.0;
			this.m_tickTime			= 0.0;
		}
		
		/**
		 * Resets the anim object so it can be reused
		 */
		public function reset():void 
		{			
			// soft reset first
			this.softReset();
			
			// return our params to the anim pool
			if ( this.m_params != null )
				AnimPool.returnAnimParams( this.m_params );
				
			// if we have a prev, reset the links
			if ( this.prev != null )
				this.prev.next = this.next;
				
			// if we have a next, reset the links
			if ( this.next != null )
				this.next.prev = this.prev;
				
			// reset any bools
			this.isIdleAnim				= false;
			this.m_calledLoadGraphics	= false;
				
			// null all our objects
			this.prev					= null;
			this.next					= null;
			this.m_params				= null;
			this.m_frames				= null;
			this.m_target				= null;
			this.m_dObjAdd				= null;
			this.m_dObjAddInstance		= null;
			this.m_dObjAddInstancePos	= null;
			this.m_dObjAddTo			= null;
			this.m_dObjAddToPos			= null;
		}
		
		/**
		 * Updates the Anim. Animations will play based on the tps set initially
		 * @param dt The time since the last update
		 */
		public function update( dt:Number ):void
		{
			// if we're paused or finished, do nothing
			if ( this.m_isPaused || this.m_isFinished )
				return;
				
			// we should have a target and a params object
			if ( this.m_target == null || this.m_params == null )
				return;
				
			// Check if we've looped
			var numArgs:int	= 0;
			var args:Array 	= null;
			if ( this.m_currFrame < this.m_prevFrameTick && this.m_params.onLoop != null && !this.isIdleAnim )
			{
				// concat up our args, and call our function (figure out if we're passing the anim or not)
				numArgs	= ( this.m_params.onLoopParams != null ) ? this.m_params.onLoopParams.length : 0;
				args 	= ( this.m_params.onLoop.length > numArgs ) ? [this] : null;
				if ( this.m_params.onLoopParams != null )
					args = ( args == null ) ? this.m_params.onLoopParams : args.concat( this.m_params.onLoopParams );
				this.m_params.onLoop.apply( null, args );
			}
			
			// on the off chance that we stop the anim in onLoop
			if ( this.m_params == null )
				return;
				
			// Check if we're finished the animation (as in we've left the anim up for at least one frame):
				
			// check if we're finished the animation
			if ( ( this.m_currFrame == this.m_prevFrame ) && ( this.m_currFrame == this.m_params.end ) && ( this.m_params.loops == 0 ) )
			{
				// we're finished the animation
				if ( this.m_params.onAnimFinish != null && !this.isIdleAnim && !this.m_isFinished )
				{
					// concat up our args, and call our function (figure out if we're passing the anim or not)
					numArgs	= ( this.m_params.onAnimFinishParams != null ) ? this.m_params.onAnimFinishParams.length : 0;
					args 	= ( this.m_params.onAnimFinish.length > numArgs ) ? [this] : null;
					if ( this.m_params.onAnimFinishParams != null )
						args = ( args == null ) ? this.m_params.onAnimFinishParams : args.concat( this.m_params.onAnimFinishParams );
					this.m_params.onAnimFinish.apply( null, args );
				}
				
				// if we're paused, then return (don't finish yet)
				// also check for the off chance that we stop the anim in onAnimFinish
				if ( this.m_isPaused || this.m_params == null )
					return;
					
				// set that we're finished and return
				this._setAsFinished();
				return;
			}
			
			// Check if we've moved onto a new frame:
				
			// if we've changed to a new frame, call our onUpdate + onFrame functions
			if ( this.m_currFrame != this.m_prevFrameTick )
			{				
				// play the anim
				if ( this.m_target is MovieClip )
					( this.m_target as MovieClip ).gotoAndStop( this.m_currFrame );
				else
				{
					// set the right bitmapdata
					if ( this.m_currFrame < this.m_params.bmds.length )
						( this.m_target as Bitmap ).bitmapData = this.m_params.bmds[this.m_currFrame];
						
					// set any offset (take scale into account)
					this._setPos( true );
				}
				
				// keep track of the x, y and scale params if needed as they might have changed (so when resuming/copying, the right params are used)
				// make sure to remove the offset if this is a bitmap anim otherwise your graphics will move over time on their own
				this._setPos( false );
				if ( !this._isNaN( this.m_params.scaleX ) ) this.m_params.scaleX = this.m_target.scaleX;
				if ( !this._isNaN( this.m_params.scaleY ) ) this.m_params.scaleY = this.m_target.scaleY;
			
				// onUpdate:
					
				// if we have an onUpdate callback, call it
				if ( this.m_params.onUpdate != null )
				{
					// concat up our args, and call our function (figure out if we're passing the anim or not)
					numArgs	= ( this.m_params.onUpdateParams != null ) ? this.m_params.onUpdateParams.length : 0;
					args 	= ( this.m_params.onUpdate.length > numArgs ) ? [this] : null;
					if ( this.m_params.onUpdateParams != null )
						args = ( args == null ) ? this.m_params.onUpdateParams : args.concat( this.m_params.onUpdateParams );
					this.m_params.onUpdate.apply( null, args );
				}
				
				// on the off chance that we stop the anim in onUpdate
				if ( this.m_params == null )
					return;
			
				// onFrame:
				
				// if we have an onFrame callback, check if we're on or passed any frames, and call it
				if ( this.m_params.onFrame != null && !this.isIdleAnim && this.m_frames != null )
				{
					// go through our frames, and check if we've passed any (or are on one)
					var len:int = this.m_frames.length;
					for ( var i:int = 0; i < len; i++ )
					{
						// get our frame
						var frame:AnimFrameInfo = this.m_frames[i];
						
						// ignore the default frame
						if ( frame.name == Anim.DEFAULT )
							continue;
							
						// if we're playing a sub anim (using an AnimFrameInfo), ignore that one as well
						if ( this.m_params.animFrameInfo != null && frame.name == this.m_params.animFrameInfo.name )
							continue;
							
						// if we're looping to a specific frame, ignore that one as well
						if ( this.m_params.loopTo != null && frame.start == this.m_params.loopTo )
							continue;
						
						// our check is different depending on if we looped or not
						if ( this.m_currFrame >= this.m_prevFrame )
						{
							// it's a normal incret, check from prev to curr
							if ( frame.start > this.m_prevFrame && frame.start <= this.m_currFrame )
							{
								args = this._getOnFrameParams( frame );
								this.m_params.onFrame.apply( null, args );
							}
						}
						else
						{
							// we've looped, so check from prev to end, then start to curr
							if ( ( frame.start > this.m_prevFrame && frame.start <= this.m_params.end ) || ( frame.start >= this.m_params.start && frame.start <= this.m_currFrame ) )
							{
								args = this._getOnFrameParams( frame );
								this.m_params.onFrame.apply( null, args );
							}
						}
					} // end frames loop
				} // end if onFrame
				
				// on the off chance that we stop the anim in onFrame
				if ( this.m_params == null )
					return;
					
			} // end if changed frames
			
			// hold the previous frame in the last tick (so that if we're playing the anim slowly, we don't set the
			// frame, or fire the onUpdate and onFrame callbacks anymore)
			this.m_prevFrameTick = this.m_currFrame;
			
			// Move onto the next frame:
			
			// if our tps is matched
			if ( this.m_params.tps <= 0 || this.m_params.tps == DS.tps )
			{
				// hold the previous frame and move on
				this.m_prevFrame = this.m_currFrame;
				this._nextFrame();
				return;
			}
			else
			{
				// get the current time for our move
				this.m_currTime += dt;
				
				// hold the previous frame
				if ( this.m_currTime > this.m_tickTime )
					this.m_prevFrame = this.m_currFrame;
					
				// get our new frame based on our current time vs how long we're meant to be in each frame
				while ( this.m_currTime > this.m_tickTime )
				{
					this.m_currTime -= this.m_tickTime;
					this._nextFrame();
				}
			}
		}
		
		/**
		 * Chains another animation after this one (or before if this is the idle anim)
		 * @param params The params to create our animation. Can be an AnimParams object, or a normal object
		 * @param isIdle Is this an idle animation? Idle animations are special animations in that they're always last, can be interrupted, 
		 * 				don't call callbacks, and you can only have one in a list
		 * @return The new animation, or this animation if one wasn't created
		 */
		public function play( params:*, isIdle:Boolean = false ):Anim
		{
			// try and get our anim
			var anim:Anim = AnimPool.getAnim( isIdle, params, this.m_params );
			if ( anim == null )
				return this;
			
			// if this is an idle anim (then we're at the end), then we add before
			if ( this.isIdleAnim )
			{
				this._addBefore( anim );
				return anim; // return the new anim
			}
			
			// go along the list and add it in the right place
			var curr:Anim = this;
			while ( curr.next != null && !curr.next.isIdleAnim )
				curr = curr.next;
			
			// add the anim to the end and return
			if ( curr.next != null ) // if there's an idle animation, fix the links
			{
				// if it's an idle anim, and this is an idle animation, destroy the old one
				if ( anim.isIdleAnim )
					curr.next.destroy();
				else
				{
					curr.next.prev 	= anim;
					anim.next		= curr.next;
				}
			}
			
			// set the links
			curr.next 	= anim;
			anim.prev	= curr;
			
			// it worked, return the new anim
			return anim;
		}
		
		/**
		 * Adds another animation directly after this one. If this animation is the idle animation, the new animation will play
		 * directly
		 * @param params The params to create our animation. Can be an AnimParams object, or a normal object
		 * @return The new animation, or this animation if one wasn't created
		 */
		public function playNext( params:* ):Anim
		{
			// try and get our anim
			var anim:Anim = AnimPool.getAnim( false, params, this.m_params );
			if ( anim == null )
				return this;
				
			// add our anim right after this one, or if this is the idle anim,
			// then before this one
			if ( this.isIdleAnim )
				this._addBefore( anim );
			else
			{
				// fix the next link
				if ( this.next != null )
				{
					this.next.prev 	= anim;
					anim.next		= this.next;
				}
				
				// set the links between this and the anim
				this.next	= anim;
				anim.prev	= this;
			}
			
			// return the new anim
			return anim;
		}
		
		/**
		 * Pauses the Anim
		 */
		public function pause():void
		{
			this.m_isPaused = true;
		}
		
		/**
		 * Resumes the Anim. You can only resume the anim if it has a target (it might be waiting on graphics), and
		 * there's no problem with the anim (it has a params object etc)
		 */
		public function resume():void
		{
			// if we don't have any params, or a target, do nothing
			if ( this.m_params == null || this.m_target == null )
				return;
				
			// unpause us
			this.m_isPaused = false;
			
			// if we're unpausing and we're finished, set us as so
			if ( this.m_currFrame == this.m_params.end && this.m_params.loops == 0 )
				this._setAsFinished();
		}
		
		/**
		 * Skips the Anim. If this is the idle anim, then all the anims for the target are stopped
		 */
		public function skip():void
		{
			// if we're not the current anim, just fix the links and return the anim
			if ( !this.m_isCurrent )
			{
				// fix our prev link
				if ( this.prev != null )
					this.prev.next = this.next;
					
				// fix our next link
				if ( this.next != null )
					this.next.prev = this.prev;
					
				// return the anim
				AnimPool.returnAnim( this );
			}
			else
			{
				// as this is the current anim, we can just set ourselves as finished
				// and the AnimManager will take care of the rest
				// - NOTE: if this anim is the idle anim, then it won't fire the onFinish
				// function, as we don't set that as false until after this (desired behaviour)
				// - NOTE: we make a check here to make sure we're not already finished - 
				// to stop someone calling anim.skip() in the onFinish callback
				if ( !this.m_isFinished )
					this._setAsFinished();
				
				// make sure we're not idle (so that it gets returned to the pool)
				this.isIdleAnim = false;
			}
		}
		
		/**
		 * Moves an anim to a specific position
		 * @param x The x position of the anim
		 * @param y The y position of the anim
		 * @return
		 */
		public function moveTo( x:Number = 0.0, y:Number = 0.0 ):Anim
		{
			// make sure we have params
			if ( this.m_params == null )
			{
				DS.warn( this, "Can't move anim to position (" + x + ", " + y + ") as we don't have a params object" );
				return this;
			}
			
			// set our params (so if other anims copy them, they get the right one)
			this.m_params.x = x;
			this.m_params.y = y;
			
			// if we're the current anim, move us
			if ( this.m_isCurrent && this.m_target != null )
				this._setPos( true );
			
			// return the anim so we can chain
			return this;
		}
		
		/**
		 * Adds our target to a parent display object
		 * @param parent The DisplayObjectContainer that we want to add our target to
		 * @param x The x position to add us at
		 * @param y The y position to add us at
		 * @return This Anim object
		 */
		public function addToParent( parent:DisplayObjectContainer, x:Number = NaN, y:Number = NaN ):Anim
		{
			// NOTE: ( x == x ) is a fast check to see if a number is value (not NaN)
			
			// copy the params over so if the anim is copied, we get the right info
			if ( this.m_params != null )
			{
				this.m_params.parent = parent;	
				if ( !this._isNaN( x ) ) this.m_params.x = x;
				if ( !this._isNaN( y ) ) this.m_params.y = y;
			}
			
			// if we have a target, add it to the parent, otherwise
			if ( this.m_target != null && this.m_isInited )
			{				
				// set our position
				this._setPos( true );
				
				// add it
				this.m_dObjAddTo.addChild( this.m_target );
			}
			else
			{
				this.m_dObjAddTo = parent;
				if ( !this._isNaN( x ) || !this._isNaN( y ) )
					this.m_dObjAddToPos = new Point( x, y ); // store the position
			}
				
			// return the anim so we can chain
			return this;
		}
		
		/**
		 * Adds a display object to our target
		 * @param child The child DisplayObject that we want to add
		 * @param instance The instance name (if any) of a sub clip in our target
		 * @param x The x position to add us at
		 * @param y The y position to add us at
		 * @return This Anim object
		 */
		public function addChild( child:DisplayObject, instance:String = null, x:Number = NaN, y:Number = NaN ):Anim
		{
			// NOTE: ( x == x ) is a fast check to see if a number is value (not NaN)
			
			// if we have a target, and it's good, add it
			if ( this.m_target != null && this.m_target is DisplayObjectContainer && this.m_isInited )
			{
				var parent:DisplayObjectContainer = this.m_target as DisplayObjectContainer;
				
				// position the child
				if( !this._isNaN( x ) )
					child.x = x;
				if( !this._isNaN( y ) )
					child.y = y;
				
				// if we have an instance, check for it
				if ( instance != null )
				{
					// look for our instance
					var sub:DisplayObject = parent.getChildByName( instance );
					if ( sub != null && ( sub is DisplayObjectContainer ) )
						( sub as DisplayObjectContainer ).addChild( child );
					else if ( sub != null )
						DS.error( this, "Can't add child " + child + " to animation target " + this.m_target + " as while we have an object called '" + instance + "', it's not a DisplayObjectContainer" );
					else
						DS.error( this, "Can't add child " + child + " to animation target " + this.m_target + " as we couldn't find the sub clip called '" + instance + "'" );
				}
				else
					parent.addChild( child ); // just add it as normal
			}
			else if ( this.m_target != null && this.m_isInited )
				DS.error( this, "Can't add child " + child + " to animation target " + this.m_target + " as it's not a DisplayObjectContainer" );
			else
			{
				// store the child, instance name and position
				this.m_dObjAdd 			= child;
				this.m_dObjAddInstance	= instance;
				if ( !this._isNaN( x ) || !this._isNaN( y ) )
					this.m_dObjAddInstancePos = new Point( x, y );
			}
			
			// return the anim so we can chain
			return this;
		}
		
		/**
		 * Returns the AnimFrameInfo object for a specific name
		 * @param name The name of the anim that we're looking for
		 * @return The AnimFrameInfo object, or null if we can't find it
		 */
		public function getFrameInfo( name:String ):AnimFrameInfo
		{
			// no frames = no AnimFrameInfo
			if ( this.m_frames == null )
				return null;
				
			// go through our frames
			var len:int = this.m_frames.length;
			for ( var i:int = 0; i < len; i++ )
			{
				if ( this.m_frames[i].name == name )
					return this.m_frames[i];
			}
			
			// couldn't find it
			return null;
		}
		
		/**
		 * Adds a frame animation/trigger to our animation
		 * @param name The name of our frame
		 * @param start The starting frame of the animation
		 * @param end The ending frame of the animation
		 */
		public function addFrameInfo( name:String, start:int = 0, end:int = 0 ):void
		{
			// create our anim frame info object
			var info:AnimFrameInfo = new AnimFrameInfo( name, start, end );
			
			// if we don't have our vector, create it
			if ( this.m_frames == null )
			{
				// our frames will only be null if we couldn't find any (e.g. bitmap anim), or we haven't loaded in yet
				// (swf + instance), so create and save them so that other anims with the same key can recover it (or combine)
				this.m_frames = new Vector.<AnimFrameInfo>;
				if ( this.m_params.animFrameKey != null )
					AnimFrameFinder.saveFrames( this.m_params.animFrameKey, this.m_frames );
			}
			this.m_frames.push( info );
		}
		
		/**
		 * Converts the anim to a String
		 */
		public function toString():String
		{
			// check for a null params object
			var id:String 	= ( this.m_params != null ) ? this.m_params.id : null;
			var start:int	= ( this.m_params != null ) ? this.m_params.start : 0;
			var end:int		= ( this.m_params != null ) ? this.m_params.end : 0;
			return "[Anim target: " + this.m_target + ", parent: " + this.parent + ", id: " + id + ", isIdle: " + this.isIdleAnim + ", curr frame: " + this.m_currFrame + " (" + start + "-" + end + "), isPaused: " + this.m_isPaused + ", hasPrev: " + ( this.prev != null ) + ", hasNext: " + ( this.next != null ) + "]";
		}
		
		/*************************************************************************************************************/
		
		/**
		 * Sets the Anim as the current anim. Only really used for the first anim created in a chain. This is not
		 * a function that can be called from external
		 */
		internal function setAsCurrent():void
		{				
			this.m_isCurrent = true;
		}
		
		/*************************************************************************************************************/
		
		// adds an anim before the idle anim
		private function _addBefore( anim:Anim ):void
		{
			// this function is only called when we're adding an animation "after" an idle, which means it
			// gets pushed before this anim. There's a few things to look out for:
			// *	if any of our previous anims are idle (called multiple before the next frame), then we
			// 		have to add the anim before those, otherwise the links will be wrong
			// *	if this is the current anim, then we need to set us as finished, and it'll get swapped
			// 		out in the next frame.
			// *	if there's multiple idle anims before this, then destroy all but the last one
			
			// fix our previous link if we have one
			if ( this.prev != null )
			{
				// if the previous anim is an idle
				if ( this.prev.isIdleAnim )
				{
					// if we're adding a new idle, kill the previous one
					if ( anim.isIdleAnim )
					{
						var prevAnim:Anim = this.prev;
						
						// set the links on the objects linked to prev
						if ( prevAnim.prev != null )
							prevAnim.prev.next = anim;
						if ( prevAnim.next != null )
							prevAnim.next.prev = anim;
							
						// set the links on the anim
						anim.prev = prevAnim.prev;
						anim.next = prevAnim.next; // prevAnim.next = this
						
						// destroy the previous
						prevAnim.next = prevAnim.prev = null; // clear these, or the links we made will be screwed in destroy()
						prevAnim.destroy(); // destroy() on prevAnim and not this.prev, as that's changed at this stage
					}
					else
					{
						// the previous anim is idle, but this one isn't, so add it before the previous one
						
						// set the links on the object before our previous anim
						if ( this.prev.prev != null )
						{
							anim.prev			= this.prev.prev;
							this.prev.prev.next = anim;
						}
						
						// set the links on the anim
						this.prev.prev 	= anim;
						anim.next		= this.prev;
					}
						
					// if we're the current, set as finished
					if ( this.m_isCurrent )
						this._setAsFinished();
					return;
				}
				
				// prev is not an idle anim so add this one after that
				this.prev.next	= anim;
				anim.prev		= this.prev;
			}
			
			// fix the links between the anim and this one
			anim.next	= this;
			this.prev	= anim;
			
			// if this is the current anim, set us as finished
			if ( this.m_isCurrent )
				this._setAsFinished();
		}
		
		// move to the next frame
		private function _nextFrame():void
		{
			this.m_currFrame++; // advance along
			
			// if we're past the end
			if ( this.m_currFrame > this.m_params.end )
			{
				// if we're loop, set us back to the start
				if ( this.m_params.loops != 0 )
				{
					// set us back to the start
					this.m_currFrame = ( this.m_params.loopTo != null ) ? this.m_params.loopTo : this.m_params.start;
						
					// remove a loop
					if ( this.m_params.loops > 0 ) // -1 = infinite looping
						this.m_params.loops--;
				}
				else
					this.m_currFrame = this.m_params.end; // just set to the end
			}
		}
		
		// returns the parameters that we're going to pass to the onFrame callback
		private function _getOnFrameParams( frame:AnimFrameInfo ):Array
		{
			// concat up our args, and call our function (figure out if we're passing the anim or not - frame is always passed)
			var numArgs:int	= ( this.m_params.onFrameParams != null ) ? this.m_params.onFrameParams.length : 0;
			var args:Array 	= ( this.m_params.onFrame.length > ( numArgs + 1 ) ) ? [this, frame.name] : [frame.name];
			if ( this.m_params.onFrameParams != null )
				args = args.concat( this.m_params.onFrameParams );
			return args;
		}
		
		// sets that we're finished and checks if this is the last animation in our chain
		private function _setAsFinished():void
		{
			this.m_isFinished = true;
			
			// if we're finished this chain of animations:
			if ( this.m_params.onFinish != null && !this.isIdleAnim && ( this.next == null || this.next.isIdleAnim ) )
			{
				// concat up our args, and call our function (figure out if we're passing the anim or not)
				var numArgs:int	= ( this.m_params.onFinishParams != null ) ? this.m_params.onFinishParams.length : 0;
				var args:Array	= ( this.m_params.onFinish.length > numArgs ) ? [this] : null;
				if ( this.m_params.onFinishParams != null )
					args = ( args == null ) ? this.m_params.onFinishParams : args.concat( this.m_params.onFinishParams );
				this.m_params.onFinish.apply( null, args );
			}
		}
		
		// called when the swf holding our graphics has loaded
		private function _onSWFLoad( a:Asset ):void
		{			
			// in the off chance that we've skipped this anim while waiting for our graphics to load, m_params will
			// be null, so just do nothing. also on the off chance that we load, recycle the anim object, and reused
			// it before the first swf is back, check the url against our swf
			if ( this.m_params == null || this.m_params.swf != a.url )
				return;
			
			// try and find our class
			this.m_params.targetClass = a.findClass( this.m_params.instance );
			if ( this.m_params.targetClass == null )
				this._onSWFFail();
			else
				// make sure we have our frames for this animation (combine with any we've added deliberately)
				this.m_frames = AnimFrameFinder.findFrames( this.m_params.targetClass, this.m_params.animFrameEndTag, this.m_params.animFrameKey, true );
		}
		
		// called when our swf fails to load
		private function _onSWFFail():void
		{			
			// skip this one
			DS.warn( this, "Couldn't load our graphics '" + this.m_params.instance + "' from swf '" + this.m_params.swf + "' (swf load failed) so skipping this anim" );
			this.skip();
		}
		
		// sets the x and y position for either the target or the params
		private function _setPos( isTarget:Boolean ):void
		{
			// no target or params, do nothing
			if ( this.m_target == null || this.m_params == null )
				return;
				
			// the offset for bitmap data animations			
			var offX:Number = 0.0;
			var offY:Number = 0.0;
			
			// calculate the offsets (take scale into account)
			if ( this.m_params.bmdOffsets != null && this.m_currFrame < this.m_params.bmdOffsets.length )
			{
				// Check if the object exist
				var p:Point = this.m_params.bmdOffsets[this.m_currFrame];
				if ( p == null )
					DS.error( this, "[setPos] Try to get offset for frame (" + this.m_currFrame + ") but there is no point for it" );
				else
				{
					// Set offset
					offX = p.x * ( !this._isNaN( this.m_params.scaleX ) ? this.m_params.scaleX : 1.0 );
					offY = p.y * ( !this._isNaN( this.m_params.scaleY ) ? this.m_params.scaleY : 1.0 );
				}
			}
			
			// do we have a position in our params?
			var isSetX:Boolean = !this._isNaN( this.m_params.x );
			var isSetY:Boolean = !this._isNaN( this.m_params.y );
			
			// if our params positions are null, set either the target or the params
			if ( isSetX || offX != 0.0 )
			{
				if ( isTarget )
					this.m_target.x = ( ( isSetX ) ? this.m_params.x : 0.0 ) + offX;
				else if ( isSetX )
					this.m_params.x = this.m_target.x - offX;
			}
			if ( isSetY || offY != 0.0 )
			{
				if ( isTarget )
					this.m_target.y = ( ( isSetY ) ? this.m_params.y : 0.0 ) + offY;
				else if ( isSetY )
					this.m_params.y = this.m_target.y - offY;
			}
		}
		
		// helper function to determine if a number if NaN
		private function _isNaN( n:Number ):Boolean
		{
			// ( n != n ) will be true if the number is NaN
			return ( n != n );
		}
		
	}

}