package com.divillysausages.sprint.anim 
{
	import com.divillysausages.ds.DS;
	import flash.display.FrameLabel;
	import flash.display.MovieClip;
	import flash.utils.Dictionary;
	import flash.utils.getDefinitionByName;
	import flash.utils.getQualifiedClassName;
	
	/**
	 * The class that will go through a clip and remove all the frame info from inside it
	 * @author Damian Connolly
	 */
	internal class AnimFrameFinder 
	{
		
		/*************************************************************************************************************/
		
		private static var m_frames:Dictionary = new Dictionary; // the frames info for the different movieclips
		
		/*************************************************************************************************************/
		
		/**
		 * Gets the frames for a particular movieclip
		 * @param obj The MovieClip or class for the movieclip that we're getting the frames for (can be null if using animFrameKey)
		 * @param animEndTag If the anims are going from one frame to another, this is the end tag to determine the "end" frame
		 * @param animFrameKey The key to save these AnimFrameInfos under (to stop from looking for them again). Useful if we have
		 * multiple animations under the same class type (e.g. [object MovieClip])
		 * @param combine Should we combine any frame info that we find with any existing frames from our frame key
		 * @return The Vector of AnimFrameInfo, or null if mc is wrong, or if there's no frames in the movieclip
		 */
		public static function findFrames( obj:* = null, animEndTag:String = null, animFrameKey:String = null, combine:Boolean = false ):Vector.<AnimFrameInfo>
		{
			// make sure we either have an obj or an animFrameKey
			if ( obj == null && animFrameKey == null )
			{
				DS.error( AnimFrameFinder, "Can't find frames as neither an object or an animFrameKey was passed" );
				return null;
			}
			
			// if our animFrameKey isn't null and we have this in our dictionary, return it
			if ( !combine && animFrameKey != null && animFrameKey in AnimFrameFinder.m_frames )
				return AnimFrameFinder.m_frames[animFrameKey];
			
			// make sure our obj is good
			if ( obj != null && !( obj is MovieClip ) && !( obj is Class ) )
			{
				DS.error( AnimFrameFinder, "Can't find the frames for " + obj + " as it's not a MovieClip or a Class" );
				return null;
			}
			
			// check if we have a valid movieclip
			var mc:MovieClip = ( obj == null ) ? null : ( obj is MovieClip ) ? obj : ( ( new obj ) as MovieClip );
			if ( obj != null && mc == null )
			{
				DS.error( AnimFrameFinder, "Can't find the frames for " + obj + " as it's not a MovieClip or a class for a MovieClip" );
				return null;
			}
			
			// if the mc is null at this point, then we can't find any frames (it's not a movieclip)
			if ( mc == null )
				return null;
			
			// find the class of the object
			var clazz:Class = ( obj is Class ) ? obj : getDefinitionByName( getQualifiedClassName( mc ) ) as Class;
			
			// if we already have this class in our dictionary, return it
			if ( !combine && clazz != null && clazz in AnimFrameFinder.m_frames )
				return AnimFrameFinder.m_frames[clazz];
				
			// we don't have any frame info, go find it
			mc.stop();
			
			// find the vector of anim info objects
			var v:Vector.<AnimFrameInfo> = AnimFrameFinder._findFramesInMC( mc, animEndTag );
			
			// if we should combine
			if ( combine )
			{
				// do we have a vector currently that we can combine to
				var orig:Vector.<AnimFrameInfo> = ( animFrameKey != null && animFrameKey in AnimFrameFinder.m_frames ) ? AnimFrameFinder.m_frames[animFrameKey] : 
													( clazz in AnimFrameFinder.m_frames ) ? AnimFrameFinder.m_frames[clazz] : 
														null;
				
				// if we have some anim frame infos already..
				if ( orig != null )
				{
					// concat the original with the new and return that as other anims might be currently using that vector
					orig.concat( v );
					return orig;
				}
			}
			
			// store it (even if it's null, it stops us looking a second time)
			if ( animFrameKey != null )
				AnimFrameFinder.m_frames[animFrameKey] = v;
			else
				AnimFrameFinder.m_frames[clazz] = v;
			return v;
		}
		
		/**
		 * Saves a Vector of AnimFrameInfo objects user a specific key
		 * @param frameKey The frameKey that we're using to save this
		 * @param frames The Vector of AnimFrameInfo objects that we want to save
		 */
		public static function saveFrames( frameKey:String, frames:Vector.<AnimFrameInfo> ):void
		{
			// don't overwrite if it's already there
			if ( frameKey in AnimFrameFinder.m_frames )
				return DS.error( AnimFrameFinder, "Can't save AnimFrameInfo vector under key '" + frameKey + "' as there's already something there" );
				
			// save it
			AnimFrameFinder.m_frames[frameKey] = frames;
		}
		
		/*************************************************************************************************************/
		
		/**
		 * Finds all the frames in a MovieClip
		 * @param mc The MovieClip that we're looking at
		 * @param animEndTag If we want to take anims as going from one frame to another, this is the end tag
		 * @return A Vector of AnimFrameInfo objects
		 */
		private static function _findFramesInMC( mc:MovieClip, animEndTag:String = null ):Vector.<AnimFrameInfo>
		{			
			var a:Array 		= mc.currentLabels;	// the array of FrameLabels
			var len:int			= a.length;			// the length of the array
			var f:FrameLabel	= null;				// the current frame label
			var nf:FrameLabel	= null;				// the next frame label
			var sf:FrameLabel	= null;				// the starting label (for when matching animations by {framelabel} - {framelabel + this.m_animEndTag}
			var start:int		= 0;				// the start of the anim
			var end:int			= 0;				// the end of the anim
				
			// create our vector
			var v:Vector.<AnimFrameInfo> = new Vector.<AnimFrameInfo>;
			
			// if there's no labels, set up the default
			if ( a.length == 0 )
			{
				v.push( new AnimFrameInfo( Anim.DEFAULT, 1, mc.totalFrames ) );
				return v;
			}
			
			// go through our frame labels
			for ( var i:int = 0; i < len; i++ )
			{
				f = a[i];	// hold the current FrameLabel
				
				// the default will run from frame 1 to the first frame tag (if it's not on frame 1)
				if ( f.frame > 1 && v.length == 0 && sf == null )
					v.push( new AnimFrameInfo( Anim.DEFAULT, 1, f.frame - 1 ) );
				
				// if each animation runs from one frame to the next
				if ( animEndTag == null )
				{
					nf 		= ( i < len - 1 ) ? a[i + 1] : null; 				// get the next label
					start 	= f.frame;											// get the starting frame
					end		= ( nf != null ) ? nf.frame - 1 : mc.totalFrames;	// get the end frame
					
					// create and store the anim info
					v.push( new AnimFrameInfo( f.name, start, end ) );
				}
				else
				{
					// if we don't have a starting frame yet
					if ( sf == null )
					{
						sf 		= f;		// hold the starting FrameLabel
						start 	= sf.frame;	// get the starting frame
						
						// if this is the last frame label, then we use the end
						if ( i == len - 1 )
						{
							end = mc.totalFrames;
							v.push( new AnimFrameInfo( sf.name, start, end ) );
						}
					}
					else
					{
						// check does this frame label match the starting label + the end tag
						if ( f.name == sf.name + animEndTag )
						{
							// get the end frame, create the AnimFrameInfo, and clear the starting frame var
							end = f.frame;
							v.push( new AnimFrameInfo( sf.name, start, end ) );
							sf	= null;
						}
					}
				}
			}
			
			// return our vector
			return v;
		}
		
	}

}