package com.divillysausages.sprint.anim 
{
	import com.divillysausages.ds.core.IUpdateObj;
	import com.divillysausages.ds.DS;
	
	/**
	 * The AnimManager is the class that controls all the animations that go through it
	 * @author Damian Connolly
	 */
	public class AnimManager implements IUpdateObj
	{
		
		/*************************************************************************************************************/
		
		private static var m_creating:Boolean 		= false;	// are we creating the manager?
		private static var m_instance:AnimManager	= null;		// the singleton instance of this
		
		/*************************************************************************************************************/
		
		/**
		 * The singleton instance of the AnimManager
		 */
		public static function get instance():AnimManager
		{
			if ( AnimManager.m_instance == null )
			{
				AnimManager.m_creating = true;
				AnimManager.m_instance = new AnimManager;
				AnimManager.m_creating = false;
			}
			return AnimManager.m_instance;
		}
		
		/*************************************************************************************************************/
		
		private var m_anims:Object 		= null; // the animations that we're currently controlling
		private var m_active:Boolean	= true;	// is the anim manager active?
		
		/*************************************************************************************************************/
		
		/**
		 * Is the AnimManager active?
		 */
		public function get active():Boolean { return this.m_active; }
		
		/*************************************************************************************************************/
		
		/**
		 * Creates the AnimManager. As it's a singleton, you should use the instance static property
		 */
		public function AnimManager() 
		{
			if ( !AnimManager.m_creating )
				throw new Error( "AnimManager is a singleton. Please use the instance static getter" );
				
			// create our animations object
			this.m_anims = new Object;
			
			// add the anim manager to the update
			DS.addToUpdate( this );
		}
		
		/**
		 * Updates the AnimManager
		 * @param dt The delta time between this and the last update
		 */
		public function update( dt:Number ):void
		{
			// go through all our anims and update them
			for ( var id:String in this.m_anims )
			{							
				// get our anim
				var anim:Anim = this.m_anims[id];
				
				// if this anim hasn't been inited yet, do so
				if ( !anim.isInited )
					anim.init();
				
				anim.update( dt ); // update the anim
					
				// if the anim has finished, then move to the next one. find it, and init it
				if ( anim.isFinished )
				{
					// if the anim is an idle one, then we keep it
					var next:Anim = null;
					if ( anim.isIdleAnim )
					{
						// go backwards until we get to the start of the list
						next = anim;
						while ( next.prev != null )
						{
							next = next.prev;
							
							// if this is an idle anim, then we're replacing another one, so return that one
							if ( next.isIdleAnim )
							{
								AnimPool.returnAnim( next.next );
								next.next = null;
							}
						}
							
						// soft reset
						if ( anim.isInited ) // isInited will be true if we haven't returned an anim (found another idle)
							anim.softReset();
					}
					else
					{
						next = anim.next;
						AnimPool.returnAnim( anim ); // return the old anim
					}
					
					// if we have a next, use it, otherwise remove the anim from our list
					if ( next != null )
					{
						this.m_anims[id] = next;
						
						// if this anim hasn't been inited yet, do so
						if ( !next.isInited )
							next.init();
					}
					else
					{
						this.m_anims[id] = null;
						delete this.m_anims[id];
					}
				} // end of if isFinished
			}
		}
		
		/**
		 * Gets the current animation for a specific id
		 * @param id The ID of the animation that we're looking for
		 * @param idle Are we looking for the first anim, or the idle anim
		 * @return The Anim object, or null if we can't find it
		 */
		public function getAnim( id:String, idle:Boolean = false ):Anim
		{
			// check that it's there
			if ( !( id in this.m_anims ) )
				return null;
				
			// if we're looking for the first (or the first is the idle), just return it
			var curr:Anim = this.m_anims[id];
			if ( !idle || curr.isIdleAnim )
				return curr;
				
			// follow the chain to find the idle
			while ( curr != null && !curr.isIdleAnim )
				curr = curr.next;
				
			// will either be null if there's no idle, or the idle
			return curr;
		}
		
		/**
		 * Adds an animation to the AnimManager
		 * @param id The ID of the animation that we want to add. Pass the same ID if you want to chain anims
		 * @param params The params we want to use for the animation. This can either be an normal object, or an AnimParams object
		 * @param isIdle Is this an idle anim? Idle animations are special animations in that they're always last, can be interrupted, 
		 * 				don't call callbacks, and you can only have one in a list
		 * @return The Anim object, or null if something went wrong
		 */
		public function playAnim( id:String, params:*, isIdle:Boolean = false ):Anim
		{
			// check if we already have anims for this object
			var curr:Anim = this.getAnim( id );
			if ( curr == null )
			{
				// create our anim
				var anim:Anim = AnimPool.getAnim( isIdle, params, null, true );
				if ( anim == null )
					return null; // something went wrong, and we've no other anim
					
				// if the anim doesn't have an id, set it
				if ( anim.params.id == null )
					anim.params.id = id;
					
				// add the anim to our object
				this.m_anims[id] = anim;
				return anim;
			}
				
			// add the anim to the end
			return curr.play( params, isIdle );
		}
		
		/**
		 * Stops all the chained animations for a particular ID
		 * @param id The ID of the animations that we want to stop
		 */
		public function stopAnims( id:String ):void
		{
			if ( id == null )
				return;
				
			// get the animation
			var curr:Anim = this.getAnim( id );
			if ( curr == null )
				return;
				
			// go along the list and clear all animations
			while ( curr != null )
			{
				var next:Anim = curr.next;
				AnimPool.returnAnim( curr );
				curr = next;
			}
			
			// remove the id from the project
			this.m_anims[id] = null;
			delete this.m_anims[id];
		}
		
		/**
		 * Gets an AnimParams object from our object pool
		 * @return An AnimParams object
		 */
		public function checkOutAnimParams():AnimParams
		{
			return AnimPool.getAnimParams();
		}
		
		/**
		 * Returns an AnimParams object to our object pool
		 * @param params An AnimParams object
		 */
		public function checkInAnimParams( params:AnimParams ):void
		{
			return AnimPool.returnAnimParams( params );
		}
		
	}

}