package com.divillysausages.sprint.anim 
{
	import com.divillysausages.ds.collections.objectpool.ObjectPool;
	import com.divillysausages.ds.DS;
	import com.divillysausages.ds.util.ClassUtils;
	import com.divillysausages.ds.util.ObjectUtils;
	import flash.display.MovieClip;
	
	/**
	 * An object pool that we use to get our anim objects, internal so only the AnimManager and Anim class can use it
	 * @author Damian Connolly
	 */
	internal class AnimPool 
	{
		
		/*************************************************************************************************************/
		
		private static var m_animsPool:ObjectPool 	= null; // the object pool for our anim objects
		private static var m_paramsPool:ObjectPool	= null;	// the object pool for our anim params
		
		/*************************************************************************************************************/
		
		/**
		 * Gets an anim object from our pool
		 * @param isIdle Is this an idle anim? Idle animations are special animations in that they're always last, can be interrupted, 
		 * 				don't call callbacks, and you can only have one in a list
		 * @param params The params for the animation. Can be either an AnimParams or an Object
		 * @param copyParams The AnimParams that we can use to copy common stuff like parent and position etc
		 * @param isCurrent Is this the current anim (i.e. the first one in a chain)
		 * @return The Anim object, or null if something went wrong
		 */
		public static function getAnim( isIdle:Boolean, params:*, copyParams:AnimParams = null, isCurrent:Boolean = false ):Anim
		{
			// our params need to be either an AnimParams object, or an Object
			var isObject:Boolean = ObjectUtils.isObject( params );
			if ( params == null || ( !( params is AnimParams ) && !isObject ) )
			{
				DS.error( AnimPool, "Can't get a new anim as the params (" + params + ") aren't an AnimParams or an Object" );
				return null;
			}
			
			// set up our params
			var p:AnimParams = ( params is AnimParams ) ? params : AnimPool.getAnimParams( params );
			
			// if we have an anim to copy params from, copy them over
			if ( copyParams != null )
				p.copyCommonParams( copyParams );
				
			// validate them
			if ( !p.validate() )
			{
				DS.error( AnimPool, "Can't get a new anim as the params (" + params + ") didn't validate" );
				return null;
			}
			
			// create our pool if it needs to be created
			if ( AnimPool.m_animsPool == null )
				AnimPool.m_animsPool = new ObjectPool( Anim, 50, false );
			
			// get a new anim
			var anim:Anim = AnimPool.m_animsPool.checkOut() as Anim;
			
			// set our properties
			anim.isIdleAnim = isIdle;
			anim.params		= p;
			
			// set if it's the current
			anim.setAsCurrent();
			
			// if we have a target movieclip or a target class, get the frames for it
			if ( anim.params.target != null )
			{
				// if our target isn't a movieclip and we have a animFrameKey, or if we're a movieclip, find our frames
				if ( !( anim.params.target is MovieClip ) && anim.params.animFrameKey != null )
					anim.frames = AnimFrameFinder.findFrames( null, anim.params.animFrameEndTag, anim.params.animFrameKey );
				else if( anim.params.target is MovieClip )
					anim.frames = AnimFrameFinder.findFrames( anim.params.target, anim.params.animFrameEndTag, anim.params.animFrameKey );
			}
			else if ( anim.params.targetClass != null )
			{
				// is this a movieclip class (or a subclass thereof)?
				var isMC:Boolean = ClassUtils.isSubclassOf( anim.params.targetClass, MovieClip );
				
				if ( !isMC && anim.params.animFrameKey != null )
					anim.frames = AnimFrameFinder.findFrames( null, anim.params.animFrameEndTag, anim.params.animFrameKey );
				else if( isMC )
					anim.frames = AnimFrameFinder.findFrames( anim.params.targetClass, anim.params.animFrameEndTag, anim.params.animFrameKey );
			}
			
			// return the anim
			return anim;
		}
		
		/**
		 * Returns an anim to the pool of anims
		 * @param anim The anim to return
		 */
		public static function returnAnim( anim:Anim ):void
		{
			// if our pool doesn't exist, then we can't return the anim
			if ( AnimPool.m_animsPool == null )
			{
				DS.warn( AnimPool, "Can't return the anim " + anim + " as there's no pool" );
				return;
			}
			AnimPool.m_animsPool.checkIn( anim );
		}
		
		/**
		 * Gets an anim params object from our pool
		 * @param params An object to set our parameters from
		 * @return The AnimParams object
		 */
		public static function getAnimParams( params:* = null ):AnimParams
		{			
			// create our pool if it needs to be created
			if ( AnimPool.m_paramsPool == null )
				AnimPool.m_paramsPool = new ObjectPool( AnimParams, 50, false );
			
			// check out a new AnimParams
			var p:AnimParams = AnimPool.m_paramsPool.checkOut() as AnimParams;
			
			// set our params if we have some
			if ( params != null )
				p.init( params );
				
			// return the anim params object
			return p;
		}
		
		/**
		 * Returns an AnimParams object to the pool of AnimParams
		 * @param params The AnimParams to return
		 */
		public static function returnAnimParams( params:AnimParams ):void
		{
			// if our pool doesn't exist, then we can't return the anim
			if ( AnimPool.m_paramsPool == null )
			{
				DS.warn( AnimPool, "Can't return AnimParams " + params + " as there's no pool" );
				return;
			}
			AnimPool.m_paramsPool.checkIn( params );
		}
		
		/**
		 * Destroys the anim pool and clears it for garbage collection
		 */
		public static function destroy():void
		{
			// destroy our anims pool
			if ( AnimPool.m_animsPool != null )
				AnimPool.m_animsPool.destroy();
				
			// destroy our params pool
			if ( AnimPool.m_paramsPool != null )
				AnimPool.m_paramsPool.destroy();
				
			// null our params
			AnimPool.m_animsPool 	= null;
			AnimPool.m_paramsPool	= null;
		}
		
	}

}